from typing import List, Tuple

from pydantic import BaseModel, validator
from typing_extensions import Literal

from aporia.sdk.widgets.base import BaseWidget, DisplayOptions, WidgetType


class TextWidgetDataOptionsAnnotations(BaseModel):
    text: str
    fontSize: int

    @validator("fontSize")
    def _validate_font_size(cls, value, values):
        if value > 32 or value < 8:
            raise ValueError("Font size must be in the range [8,32]")
        return value


class TextWidgetDataOptionsFilters(BaseModel):
    annotations: TextWidgetDataOptionsAnnotations


class TextWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: TextWidgetDataOptionsFilters


class TextWidget(BaseWidget):
    name: str = "Text Widget"
    type: Literal[WidgetType.TEXT] = WidgetType.TEXT
    dataOptions: List[TextWidgetDataOptions] # Length must be 1?

    @classmethod
    def create(
        cls, position: Tuple[int, int], size: Tuple[int, int], text: str, font: int = 24
    ) -> "TextWidget":
        return TextWidget(
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            dataOptions=[
                TextWidgetDataOptions(
                    filters=TextWidgetDataOptionsFilters(
                        annotations=TextWidgetDataOptionsAnnotations(text=text, fontSize=font)
                    )
                )
            ],
        )



from typing import Any, Dict, List, Optional, Tuple

from pydantic import BaseModel, Field
from typing_extensions import Literal

from aporia.sdk.datasets import DatasetType
from aporia.sdk.fields import Field, FieldGroup
from aporia.sdk.segments import Segment
from aporia.sdk.widgets.base import BaseWidget, DisplayOptions, VersionSelection, WidgetType


class HistogramOverTimeWidgetDataOptionsAnnotations(BaseModel):
    field: Dict
    phase: Literal[DatasetType.SERVING] = DatasetType.SERVING
    version: VersionSelection = VersionSelection()
    fieldCategory: FieldGroup
    granularity: Optional[str] = None
    dataSegment: Optional[Dict] = None
    dataSegmentId: Optional[str] = None
    dataSegmentName: Optional[str] = None
    dataSegmentValue: Optional[Any] = None  # Optional[Union[int, float, str]] = None
    dataSegmentBucketName: Optional[str] = None

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class HistogramOverTimeWidgetDataOptionsFilters(BaseModel):
    annotations: HistogramOverTimeWidgetDataOptionsAnnotations


class HistogramOverTimeWidgetDataOptions(BaseModel):
    display: DisplayOptions = DisplayOptions()
    filters: HistogramOverTimeWidgetDataOptionsFilters


class HistogramOverTimeWidget(BaseWidget):
    type: Literal[WidgetType.TIME_SERIES_HISTOGRAM] = WidgetType.TIME_SERIES_HISTOGRAM
    dataOptions: List[HistogramOverTimeWidgetDataOptions]  # Length must be at most 2?

    @classmethod
    def create(
        cls,
        position: Tuple[int, int],
        size: Tuple[int, int],
        title: str,
        field: Field,
        version: VersionSelection = VersionSelection(),
        segment: Optional[Segment] = None,
        segment_value: Optional[Any] = None,
        granularity: Optional[str] = None,
    ) -> "HistogramOverTimeWidget":
        if (segment is None) != (segment_value is None):
            raise ValueError(
                "For using segments, both segment and segment_value arguments are needed"
            )

        return HistogramOverTimeWidget(
            x=position[0],
            y=position[1],
            w=size[0],
            h=size[1],
            name=title,
            dataOptions=[
                HistogramOverTimeWidgetDataOptions(
                    filters=HistogramOverTimeWidgetDataOptionsFilters(
                        annotations=HistogramOverTimeWidgetDataOptionsAnnotations(
                            version=version,
                            field=field.to_widget(),
                            granularity=granularity,
                            fieldCategory=field.group,
                            dataSegment=segment.to_widget() if segment is not None else None,
                            dataSegmentId=segment.id if segment is not None else None,
                            dataSegmentName=segment.name if segment is not None else None,
                            dataSegmentValue=segment_value if segment is not None else None,
                            dataSegmentBucketName=[
                                bucket["name"]
                                for bucket in segment.to_widget()["buckets"]
                                if bucket["value"] == segment_value
                            ][0]
                            if segment is not None
                            else None,
                        ),
                    )
                )
            ],
        )



from aporia.sdk.widgets.base import BaseWidget
from aporia.sdk.widgets.data_health_table_widget import DataHealthTableWidget
from aporia.sdk.widgets.distribution_widget import DistributionWidget
from aporia.sdk.widgets.histogram_over_time_widget import HistogramOverTimeWidget
from aporia.sdk.widgets.metric_widget import MetricWidget
from aporia.sdk.widgets.text_widget import TextWidget
from aporia.sdk.widgets.timeseries_widget import TimeSeriesWidget



import datetime
from enum import Enum
from typing import Dict, Optional, Tuple
import uuid

from pydantic import BaseModel, Field, validator

# TODO: Check with python 3.8


class FixedTimeframe(BaseModel):
    to: datetime.datetime
    from_impl: datetime.datetime = Field(alias="from")

    @validator("to", "from_impl")
    def _validate_utc(cls, value, values):
        if value.tzinfo != datetime.timezone.utc:
            raise ValueError("Timezones must be in UTC")
        return value

    def dict(self, *args, **kwargs):
        return {"to": self.to.isoformat(), "from": self.from_impl.isoformat()}


class DisplayOptions(BaseModel):
    color: int = 0  # 1 for compare


class VersionSelection(BaseModel):
    id: str = "*"


class BaselineType(str, Enum):
    TRAINING = "training"
    RELATIVE_TIME_PERIOD = "relativeTimePeriod"


class TimeUnit(str, Enum):
    DAY = "day"
    MONTH = "month"


class BaselineConfiguration(BaseModel):
    type: BaselineType
    unit: Optional[TimeUnit] = None
    duration: Optional[int] = None

    @validator("unit")
    def _validate_unit(cls, value, values) -> Optional[TimeUnit]:
        if values["type"] == "training" and value is not None:
            raise ValueError("Unit can't be used with training baseline")
        return value

    @validator("duration")
    def _validate_duration(cls, value, values) -> Optional[int]:
        if values["type"] == "training" and value is not None:
            raise ValueError("Unit can't be used with training baseline")
        return value

    def dict(self, *args, **kwargs):
        return {k: v for k, v in super().dict(*args, **kwargs).items() if v is not None}


class MetricParameters(BaseModel):
    id: str


class WidgetType(str, Enum):
    TEXT = "text"
    ANOMALY_TABLE = "anomaly-table"
    DISTRIBUTION = "distribution"
    TIME_SERIES_HISTOGRAM = "time-series-histogram"
    METRIC = "metric"
    TIME_SERIES = "time-series"


class BaseWidget(BaseModel):
    i: str = Field(default_factory=lambda: str(uuid.uuid4()))
    h: int
    w: int  # Capped at 12
    x: int  # Capped at 11?
    y: int
    name: str
    type: WidgetType
    moved: bool = False
    resizable: bool = True
    dependencies: Optional[
        Dict
    ] = None  # TODO: Investigate this, and consider if you want to raise error if it has issues

    @validator("x")
    def _validate_x(cls, value, values):
        if value < 0:
            raise ValueError("x must be non-negative")
        if value >= 12:
            raise ValueError("The grid's x-axis is limited at 12 cells")
        return value

    @validator("y")
    def _validate_y(cls, value, values):
        if value < 0:
            raise ValueError("y must be non-negative")
        return value

    @validator("w")
    def _validate_w(cls, value, values):
        if value <= 0:
            raise ValueError("w must be positive")
        if value > 12:
            raise ValueError("The grid's x-axis is limited at 12 cells")
        return value

    @validator("h")
    def _validate_h(cls, value, values):
        if value <= 0:
            raise ValueError("h must be positive")
        return value

    @classmethod
    def create(
        cls, position: Tuple[int, int], size: Tuple[int, int], *args, **kwargs
    ) -> "BaseWidget":
        raise NotImplementedError()
